#!/usr/bin/python
# -*- coding: utf-8 -*-

# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2000-2022 Jens Lechtenbörger

u"""Functional dependencies (FDs) are a major tool for database design.

Introduction
============
The module
[functional_dependencies](https://gitlab.com/oer/cs/functional-dependencies/-/blob/master/functional_dependencies/functional_dependencies.py)
defines the classes `FD`, `FDSet`, and `RelSchema` to represent
a single functional dependency (FD), a set of FDs, and a relation schema,
respectively.  Here, a relation schema
is defined by a set of attributes and a set of FDs.  In particular,
`RelSchema.synthesize()` *synthesizes* a given schema into a set of relation
schemata in 3rd normal form (3NF).  Thus, the given schema does not need to
satisfy any normal form, while the set of output schemata will be in 3NF.
While the synthesis algorithm goes back to Bernstein (1976),
https://doi.org/10.1145/320493.320489,
notation and algorithms in this module follow that book:
Vossen (1999): Datenbankmodelle, Datenbanksprachen und
Datenbankmanagementsysteme, 3. Aufl., Oldenbourg

*Relation schemata* define tabular structures (in a spirit similar to CREATE
TABLE statements of SQL) with attributes (column headers with data types) and
constraints (here, we just consider functional dependencies, which generalize
primary keys).  A *relation* is an *instance* of a schema, i.e., a set of rows
or *tuples* that obeys the rules (data types and constraints) laid out by the
schema.

Functional Dependencies
=======================
An FD is an integrity constraint, restricting the set of permissible
relations/instances of a relation schema.  Intuitively, FDs capture semantic
information which attributes belong together how.
FDs are written in the form X -> Y for sets X and Y of attributes.
(In this module, attributes are just strings.  In general, each attribute has
a *domain* of permissible values, potentially including a NULL value.)
X is called *left-hand side* (lhs), Y *right-hand side* (rhs).
An FD is called *r-minimal* or *simple* if |Y| == 1, i.e., if the rhs consists
of a single element.

Following general practice, in this documentation we may represent sets of
attributes as sequences, omitting braces and commas to simplify notation.

Formally, an FD X -> Y *holds* in a relation r if no two tuples exist in r
that share the same X-value but disagree on their Y-values.
(Intuitively, each X-value functionally determines exactly one Y-value.)
An FD is *trivial* if it is satisfied in every relation.  (E.g., if Y is a
subset of X, then X -> Y is trivial: clearly, if tuples agree on the "larger"
lhs, then they also agree on the rhs.)

Instances of `FD` are constructed from lhs and rhs, each of which is
either a single attribute or set of attributes.  E.g., `fd1` below
represents the FD CustomerID -> DateOfBirth.
>>> fd1 = FD("CustomerID", "DateOfBirth")
>>> fd1.isrminimal()
True
>>> fd1.attributes() == {"CustomerID", "DateOfBirth"}
True
>>> fd2 = FD({"CustomerID", "DateOfBirth"}, {"BirthYear", "Country"})
>>> fd2.isrminimal()
False
>>> fd2.attributes() == {"CustomerID", "DateOfBirth", "BirthYear", "Country"}
True

Keys
----
A *superkey* for a relation schema R with attributes Y is a set K of attributes
such that the FD K -> Y holds.  Intuitively, the values of superkeys are
unique per relation.

The notion of superkey is merely a technical one.  We are really interested in
keys.  ("Super" does not indicate "better" but "superset of"; see next
paragraph.)

A *key* is a superkey that is minimal with respect to set inclusion.  (I.e.,
if we remove any attribute from a key, then the remaining attributes do not
functionally determine all other attributes any more.)  Thus, every superset
of a key is a superkey.  In particular, the set of all attributes is a trivial
superkey (and method `RelSchema.key()`, which determines a key for a relation
schema with its FDs, starts from that superkey to find a key).

In SQL, one of the keys of a relation schema may be declared as *primary key*.
(I do not know what a "candidate" key is.  Let me expand on that.  In his
[seminal paper on the relational
model](https://dl.acm.org/doi/10.1145/362384.362685),
Codd talks about primary keys and nonredundant primary keys, while he defines
candidate keys in his [follow-up paper on
normalization](https://www.bibsonomy.org/bibtex/24b7b528f0502ff638c837f39a3ed3732).
In the decades since then, we have come to *define* keys as nonredundant sets
of attributes.  Thus, redundant keys do not exist any longer, and
"nonredundant" stopped being a meaningful qualifier.  Moreover, I
doubt that a reasonable definition for "noncandidate" key exists, which turns
"candidate" into a meaningless qualifier.  Please do not use it unless you can
say what it means.)

In general, multiple keys may exist; thus, we talk about "a key",
not "the key" (e.g., in a table with data about student assistants, the
matriculation number, the student number, the employee number, and the tax ID
could exist as four different keys).  Note that keys are *sets* of attributes
(e.g., the  primary key of a fact table in a data warehouse contains
attributes for each of the dimensions).

An attribute that occurs in *some* key is called *key attribute* or *prime
attribute*; otherwise it is a non-key attribute (non-prime).

*Warning!*  Please be careful if you read texts on normalization or FDs that
talk about "the key".  Quite likely, they are incorrect.


Sets of FDs
===========
The class `FDSet` represents a set of FDs.
A non-simple FD X -> A1, ..., An with n>1 is equivalent to a set of n simple
FDs X -> A1, ..., X -> An (this claim requires a proof).
Such a set can be computed with `FD.rminimize()` (which also removes
trivial FDs).  As fd1 is simple and non-trivial, `rminimize()` does not have
an effect:
>>> fdset1 = fd1.rminimize()
>>> logging.debug("fdset1: %s", fdset1)
>>> len(fdset1) == 1
True
>>> fd1b = next(iter(fdset1))
>>> fd1.lhs == fd1b.lhs and fd1.rhs == fd1b.rhs
True

Non-simple FD fd2 is split into two simple FDs:
>>> fdset2 = fd2.rminimize()
>>> len(fdset2) == 2
True
>>> len([nfd for nfd in fdset2 if not nfd.isrminimal()]) == 0
True
>>> print(fdset2)
{{CustomerID, DateOfBirth} -> {BirthYear}, {CustomerID, DateOfBirth} -> {Country}}

In the following FD, DateOfBirth occurs in lhs and rhs.  Thus, `rminimize()`
removes the trivial FD {CustomerID, DateOfBirth} -> DateOfBirth.
>>> fd3 = FD({"CustomerID", "DateOfBirth"}, {"DateOfBirth", "BirthYear"})
>>> fd3mset = fd3.rminimize()
>>> len(fd3mset) == 1
True
>>> fd3 == fd3mset
False
>>> fd3b = next(iter(fd3mset))
>>> fd3 == fd3b
False
>>> fd3 != fd3b
True
>>> print(fd3b)
{CustomerID, DateOfBirth} -> {BirthYear}

Implication of FDs
------------------
Given some FDs, other (non-trivial) FDs may be *implied*, e.g., if A -> B and
B -> C hold in some relation, then also A -> C holds (this claim requires
a proof).  In this case, the FD A -> C is called *transitive*, while A -> B
and B -> C are *direct*.
>>> fds = FDSet()
>>> fds.add(FD({"CustomerID"}, {"DateOfBirth"}))
>>> fds.add(FD({"DateOfBirth"}, {"BirthYear"}))
>>> fds.isimplied({"CustomerID"}, "BirthYear")
True

Given a set F of FDs, the *closure* of F, denoted F⁺, is the set of FDs
that are implied by F.

Implication of FDs is related to the *closure* of attributes.  Given sets F of
FDs and X of attributes, the closure of X with respect to F, denoted X⁺, is
the largest set of attributes Y such that X -> Y is implied by F.
>>> fds.closure({"CustomerID"}) == {"CustomerID", "DateOfBirth", "BirthYear"}
True

FD Basis
--------
Given a set F of FDs, we may wonder what its "core" looks like, i.e., a
"small" set G of FDs such that the FDs implied by F and G are the same (i.e.,
F⁺ = G⁺).
Formally, a *basis* is such a set (note "a" basis, not "the" basis):
>>> fds2 = FDSet(fds.copy())
>>> fds2.add(FD({"CustomerID"}, {"BirthYear"}))
>>> fds2.add(FD({"CustomerID", "DateOfBirth"}, {"BirthYear"}))
>>> fds2.add(FD({"Country"}, {"Country"}))
>>> basis = fds2.basis()
>>> fds == basis and basis.basis() == basis
True

What is computed as basis here, is called "minmal cover" elsewhere (e.g.,
[Abiteboul, Hull, Vianu (1995): Foundations of Databases](https://wiki.epfl.ch/provenance2011/documents/foundations%20of%20databases-abiteboul-1995.pdf)):
Such a set G

- contains only simple FDs (guaranteed by `FD.rminimize()`),
- contains only l-minimal FDs (guaranteed by `FDSet.lminimize()`),
- is minimal or nonredundant, i.e., no subset of G has the same closure
  (tested by iterative removal attempts of individual FDs in `FDSet.basis()`).

>>> {"CustomerID"} == fds2.lminimize({"CustomerID", "DateOfBirth"}, {"BirthYear"})
True

Normal Forms
============
*Normal forms* are quality criteria for "good" schemata.  Normal forms for
relational schemata start from the first normal form (1NF), for which
different interpretations exist.  Codd initially ruled out sets as attribute
values with 1NF, while other sources rule out non-atomic values.  We don't
bother with the subtleties here but note that a person's name consisting of
components, such first and last name, is not atomic, which precludes queries
by last name and hints therefore at disadvantages of non-atomic values.
However, the boundaries are fuzzy.  Are dates including year, month, and day
atomic or not?  Are strings consisting of characters atomic?

Higher normal forms (2NF, 3NF, ...) suppose that 1NF is satisfied.  Then they
aim to avoid *redundancy* and so-called *update anomalies*, see
[Wikipedia](https://en.wikipedia.org/wiki/Database_normalization)
for examples.

3NF
---
The present Python module focuses on 3NF normalization.  Intuitively, 3NF
makes sure that each relation schema only contains attributes that belong to a
"reasonable" semantic unit.  A relation schema in 1NF is in *3NF* iff (if and
only if) for every non-key attribute A and every key K the FD K -> A is
direct.  Equivalently, it is in 3NF iff for every non-trivial FD Y -> A either
Y is a superkey or A is a key attribute.  (Note that you do not need to know
2NF to understand 3NF.  Also, we can normalize to 3NF directly, without an
intermediate step via 2NF.)

*Warning!*  Again, if some text introduces normal forms with reference to "the
key", please stay away.  (Under such approaches, it might be sufficient to add
a new attribute, say with increasing integer numbers, as primary key to
"normalize" the schema.  Clearly, adding a new attribute does not remove
redundancy and does not avoid update anomalies.  Thus, such approaches are
flawed.)

Synthesis
---------
The class of *synthesis* algorithms transforms or *normalizes* an input schema
with a given set F of FDs into a set of relation schemata in third normal form
(3NF).  The essential structure of synthesis algorithms is as follows:

1. "Minimize" F to obtain an equivalent set F’
   (remove redundancies, group common left hand sides together)
2. For each FD X → Y in F' create a relation schema over attributes X∪Y
   with *key X*
3. If no created schema contains a key for the input schema, then add a further
   schema having such a key

To "minimize", the implementation here first computes a basis for F.
Then, it creates a relation schema for each lhs occurring in the
basis.

Supplier example of Vossen (1999).
>>> fds = FDSet()
>>> fds.add(FD({"Lieferant", "Teil"}, {"Anzahl", "Ort", "Entfernung"}))
>>> fds.add(FD({"Lieferant"}, {"Ort"}))
>>> fds.add(FD({"Ort"}, {"Entfernung"}))
>>> fds.key() == {"Lieferant", "Teil"}
True
>>> lieferant = RelSchema(fds.attributes(), fds)
>>> lieferant.key() == {"Lieferant", "Teil"}
True
>>> normalisiert = lieferant.synthesize()
>>> logging.debug(_rels2string(normalisiert))
>>> len(normalisiert) == 3
True

Synthesize the account example of Vossen (1999),
which adds a global key:
>>> fds = FDSet()
>>> fds.add(FD({"Kunde"}, {"Saldo"}))
>>> fds.add(FD({"Nummer"}, {"Zweigstelle", "Saldo"}))
>>> fds.key() == {"Kunde", "Nummer"}
True
>>> kunde = RelSchema(fds.attributes(), fds)
>>> normalisiert = kunde.synthesize()
>>> logging.debug(_rels2string(normalisiert))
>>> len(normalisiert) == 3
True

Synthesize 3NF schemata for exercise 7.8 (10) of Vossen (1999).
Note that we start from 6 FDs, where one is redundant.  Which one is it why?
>>> ex10 = FDSet()
>>> ex10.add(FD({"Course"}, {"Teacher"}))
>>> ex10.add(FD({"Hour", "Room"}, {"Course"}))
>>> ex10.add(FD({"Hour", "Teacher"}, {"Room"}))
>>> ex10.add(FD({"Course", "Student"}, {"Grade"}))
>>> ex10.add(FD({"Hour", "Student"}, {"Room"}))
>>> ex10.add(FD({"Hour", "Room"}, {"Teacher"}))
>>> relschema = RelSchema(ex10.attributes(), ex10)
>>> normalized = relschema.synthesize()
>>> logging.debug(_rels2string(normalized))
>>> len(normalized) == 5
True

A schema for class usage.
>>> misdwh = FDSet()
>>> misdwh.add(FD({"AccID"}, {"AccID", "Type", "CustID"}))
>>> misdwh.add(FD({"AccID", "Date"},
... {"AccID", "Date", "Type", "CustID", "Balance", "CustAge"}))
>>> misdwh.add(FD({"AccID", "Date"}, {"Balance", "CustAge"}))
>>> misdwh.add(FD({"CustID", "Date"}, {"CustAge", "CustValue"}))
>>> misdwh.add(FD({"CustID"}, {"CustName", "DateOfBirth", "Address"}))
>>> banking = RelSchema(misdwh.attributes(), misdwh)
>>> starschema = banking.synthesize()
>>> logging.debug(_rels2string(starschema))
>>> len(starschema) == 4
True

`RelSchema.synthesize()` may output redundant schemata as an address example
of Vossen (1999) shows (a classical example of a schema in 3NF but not in
Boyce-Codd Normal Form, BCNF).
>>> fdaddr = FDSet()
>>> fdaddr.add(FD({"City", "Address"}, "ZipCode"))
>>> fdaddr.add(FD("ZipCode", "City"))
>>> addrschema = RelSchema(fdaddr.attributes(), fdaddr)
>>> addr3nf = addrschema.synthesize()
>>> logging.debug(_rels2string(addr3nf))
>>> len(addr3nf) == 2
True

Use parameter `minimize=True` to remove redundant schemata.
>>> addr3nf = addrschema.synthesize(minimize=True)
>>> logging.debug(_rels2string(addr3nf))
>>> len(addr3nf) == 1
True

A schema where keys are not immediately visible.
>>> nokey = FDSet()
>>> nokey.add(FD("A", "B"))
>>> nokeyschema = RelSchema({"A", "B", "C"}, nokey)
>>> nokey3nf = nokeyschema.synthesize()
>>> logging.debug(_rels2string(nokey3nf))
>>> len(nokey3nf) == 2
True
"""

import logging


def _set2string(elems):
    """Return string representing the set elems.

    Produce braces and commas, sort elements.
    """
    return "{{{}}}".format(", ".join(sorted([str(elem) for elem in elems])))


def _maybe2set(thing):
    """Return thing in or as set.

    If thing is a set, return unchanged.
    Otherwise, return set containing just thing.
    """
    result = thing
    if not isinstance(thing, set):
        result = set()
        result.add(thing)
    return result


def _rels2string(relations):
    """Return relations as string, one per line in random order."""
    return "\n".join([str(relation) for relation in relations])


class FD(object):
    """A functional dependency with left- and right-hand side."""

    def __init__(self, lhs, rhs):
        """Create FD with lhs and rhs.

        Each argument can be a single attribute or a set of attributes.
        """
        self.lhs = _maybe2set(lhs)
        self.rhs = _maybe2set(rhs)

    def __str__(self):
        """Return string for self.  Use sorting for unique string."""
        return "{} -> {}".format(_set2string(self.lhs), _set2string(self.rhs))

    def __eq__(self, other):
        """Return True iff lhs and rhs of self and other are equal."""
        if isinstance(other, FD):
            return self.lhs == other.lhs and self.rhs == other.rhs
        return False

    def __ne__(self, other):
        """Return not `__eq__()`."""
        return not self.__eq__(other)

    def __hash__(self):
        """Return `hash()` of string."""
        return hash(str(self))

    def isrminimal(self):
        """Return true iff self is r-minimal.

        An FD is r-minimal if the right-hand side is a singleton.
        """
        return len(self.rhs) == 1

    def attributes(self):
        """Return attributes of self (union of left- and right-hand side)."""
        return self.lhs.union(self.rhs)

    def rminimize(self):
        """Return a minimal cover of r-minimal FDs for self.

        Achieve minimality by removal of trivial FDs.
        """
        result = FDSet()
        for attr in self.rhs:
            singleton = set()
            singleton.add(attr)
            if attr in self.lhs:
                # FDs of the form XA->A are trivial.
                logging.debug("rminimize: omitted trivial FD %s",
                              FD(self.lhs, singleton))
                continue
            result.add(FD(self.lhs.copy(), singleton))
        return result


class FDSet(set):
    """A set of functional dependencies."""

    def __init__(self, fdset=None):
        """Construct FDSet, either as empty set of from given set."""
        set.__init__(self)
        if fdset:
            self.update(fdset)

    def __str__(self):
        """Return set with braces, commas, and sorted elements."""
        return _set2string(self)

    def __eq__(self, other):
        """Test for set equality of self and other."""
        return self.issubset(other) and other.issubset(self)

    def attributes(self):
        """Return set with all attributes in FDs in self."""
        result = set()
        for fdep in self:
            result.update(fdep.attributes())
        return result

    def project(self, attributes):
        """Project self to FDs that contain given attributes."""
        result = set()
        for fdep in self:
            if fdep.lhs.issubset(attributes) and fdep.rhs.issubset(attributes):
                result.add(fdep)
        return result

    def closure(self, attributes):
        """Compute closure of attributes under self."""
        result = set(attributes)
        more = True
        while more:
            more = False
            for fdep in self:
                if fdep.lhs.issubset(result) and not fdep.rhs.issubset(result):
                    more = True
                    result.update(fdep.rhs)
        return result

    def isimplied(self, lhs, rhs):
        """Return true iff lhs -> rhs is implied in self.

        Test whether the closure of lhs contains rhs, where rhs can be a
        single attribute or a set of attributes.
        """
        return self.closure(lhs).issuperset(_maybe2set(rhs))

    def lminimize(self, lhs, rhs):
        """Compute minimum subset of lhs that determines rhs.

        Repeatedly remove attributes from lhs as long as it still determines
        rhs.  Note that this removal is non-deterministic.
        """
        cand = lhs.copy()
        for attr in lhs:
            cand.remove(attr)
            if not self.isimplied(cand, rhs):
                # attr is necessary to determine rhs.  Do not remove.
                cand.add(attr)
            else:
                logging.debug(
                    "lminimize: removed %s for rhs %s: new lhs: %s",
                    attr, _set2string(rhs), _set2string(cand))
        return cand

    def key(self, attributes=None):
        """Compute (some) key based on self.

        If attributes is not None, compute for a key for those attributes;
        otherwise, compute a key for the attributes occurring in self.

        Start with entire set of attributes as superkey.  Use lminimize()
        to remove attributes as long as superkey property is satisfied.
        As lminimize() is non-deterministic, this method is non-deterministic
        as well.
        """
        if attributes is None:
            attributes = self.attributes()
        return self.lminimize(attributes, attributes)

    def basis(self):
        """Compute non-redundant r- and l-minimal basis.

        The algorithm here is based on pseudo-code on p. 166 in
        Vossen (1999): Datenbankmodelle, Datenbanksprachen und
        Datenbankmanagementsysteme, 3. Aufl., Oldenbourg
        """
        result = FDSet()
        # First, create simple FDs for self.
        for fdep in self:
            result.update(fdep.rminimize())

        # Second, make each FD l-minimal.
        rcopy = result.copy()
        for fdep in rcopy:
            lhs = self.lminimize(fdep.lhs, fdep.rhs)
            if lhs != fdep.lhs:
                result.remove(fdep)
                logging.debug(
                    "basis: replaced lhs in %s with %s",
                    fdep, _set2string(lhs))
                result.add(FD(lhs, fdep.rhs))

        # Third, remove redundant FDs.  (An FD is redundant if it is implied
        # by other FDs.)
        rcopy = result.copy()
        for fdep in rcopy:
            result.remove(fdep)
            if not result.isimplied(fdep.lhs, fdep.rhs):
                result.add(fdep)
            else:
                logging.debug("basis: removed implied FD %s", fdep)

        return result


class RelSchema(object):
    """A relation schema consists of a set of attributes and a set of FDs.

    Various normal forms exist to describe "good" schemata.  Normalization
    is the process of creating schemata that satisfy certain normal forms.
    The class of synthesis algorithms targets 3NF.
    """

    def __init__(self, attributes, fds):
        """Construct relation schema with attributes and FDs."""
        self.attributes = attributes
        self.fds = fds

    def __str__(self):
        """Return relation schema in usual pair representation."""
        return "({}, {})".format(_set2string(self.attributes),
                                 _set2string(self.fds))

    def key(self):
        """Return a key."""
        return self.fds.key(self.attributes)

    def synthesize(self, minimize=False):
        """Synthesize set of 3NF schemata for self.

        The essential step is the computation of a basis.  Then, create a
        relation schema for each lhs in the basis.  To ensure losslessness, a
        global key may need to be added.

        If `minimize` is True, remove redundant schemata from output.
        """
        basis = self.fds.basis()
        bcopy = FDSet(basis.copy())
        result = set()

        # Iterate over FDs in basis to create schemata per lhs.
        while len(basis) > 0:
            result.add(extract_by_lhs(basis, next(iter(basis))))

        # Test whether a global key is contained.
        havekey = False
        for schema in result:
            if bcopy.isimplied(schema.attributes, self.attributes):
                havekey = True
                break
        if not havekey:
            key = self.key()
            result.add(RelSchema(key, FDSet()))
            logging.debug("synthesize: added key %s", _set2string(key))

        if minimize:
            rcopy = result.copy()
            for schema in rcopy:
                for cand in rcopy:
                    if schema.attributes == cand.attributes:
                        continue
                    if schema.attributes.issubset(cand.attributes):
                        result.remove(schema)
                        logging.debug("synthesize: removed %s", schema)

        return result


def extract_by_lhs(basis, fdep):
    """Synthesize an instance of RelSchema for lhs of fdep.

    FD fdep must be an element of basis.
    We collect the lhs of fdep and all attributes that are determined by
    the lhs (of fdep and other FDs in basis that share the same lhs) into the
    result schema.

    Note: FDs with matching lhs are removed from basis.  Thus, we can
    iteratively call this until the basis is empty for a synthesis algorithm.
    """
    assert fdep in basis
    attr = set()
    attr.update(fdep.lhs)
    bcopy = FDSet(basis)
    for fdi in bcopy:
        if fdi.lhs == fdep.lhs:
            attr.update(fdi.rhs)
            basis.remove(fdi)
    result = RelSchema(attr, bcopy.project(attr))
    return result


# Run doctests by default.
if __name__ == '__main__':
    import doctest
    print(doctest.testmod())


'''
Created on 25.05.2022

@author: maxwellgerber

see https://gist.github.com/maxwellgerber/4caae07161ea66123de4d6c374387786

'''
from itertools import combinations, permutations, chain
import datetime
from functional_dependencies.fdsbase import Attribute, Set, Notation, FD
    

class FunctionalDependencySet:
    '''
    a functional dependency set
    '''

    def __init__(self, attributes:str="",title:str="",description:str="",notation:Notation=None,debug:bool=False):
        '''
        constructor
        
        Args:
            attributes(str): a string of attribute variable names of the scheme
            title(str): a title for this functional Dependency Set
            description(str): a description for this functional Dependency Set
            notation(Notation): the notation to be used
            debug(bool): if True switch debugging on
        '''
        self.title=title
        self.description=description
        if notation is None:
            notation=Notation.utf8
        self.notation=notation    
        self.debug=debug
        self.isodate=datetime.datetime.now().isoformat()
        #list of FDs of the scheme. An FD is stored as a tuple (x, y), meaning x -> y
        self.dependencies = []

        #set of attributes of the scheme
        self.attributes = Set()
        self.attribute_map={}

        self.isdecomposed = False
        
        for attr in attributes:
            self.add_attribute(attr)
            
            
    def __str__(self):
        '''
        return my text representation
        '''
        text=self.stringify_dependencies()
        return text
    
    def set_list_as_text_list(self,set_list:list,notation:Notation):
        '''
        convert a list of sets to a list of strings using the given delimiter
        
        Args:
            set_list(list): list of sets
            notation(Notation): the notation to use
            
        Returns:    
            list: of stringified sets
        '''
        text_list=[]
        for a_set in set_list: 
            text_list.append(Set.stringify_set(a_set,notation=notation))
        text_list=sorted(text_list)
        return text_list
    
    
    def stringify_dependencies(self):
        '''
        stringifies the set of dependencies
        '''
        text='{'
        delim=''
        if self.notation==Notation.math or self.notation==Notation.plain:
            fdNotation=self.notation
        else:
            fdNotation=Notation.utf8
        for left,right in self.dependencies:
            fd=FD(left,right)
            fdtext=FD.stringify_FD(fd, fdNotation)
            text+=f"{delim}{fdtext}"
            delim=","
        text+="}"
        return text
     
    def add_attribute(self, attr_var:str,attr_english_name:str=None,attr_german_name:str=None):
        '''
        add attribute to the attribute set of the scheme

        Args:
            attr_var(string): attribute variable name to be added to the scheme
            attr_english_name(string): the name of the attribute in english
            attr_german_name(string): the name of the attribute in german
        '''
        if attr_english_name is None:
            attr_english_name=attr_var
        if attr_german_name is None:
            attr_german_name=attr_english_name
        attr=Attribute(attr_var,attr_english_name,attr_german_name)
        self.attributes.add(attr_var)
        self.attribute_map[attr_var]=attr
    
    def add_dependency(self, pre, post):
        '''
        add dependency to the dependency list of the scheme

        Args:
            pre(set): attributes that initiate the FD (left of the arrow)
            post(set): attributes that are determined by the FD (right of the arrow)
        '''
        for i in chain(pre,post):
            if i not in self.attributes:
                # exception when an attribute is used that is not in the list of attributes of the dependency
                raise Exception(f"Attribute {i} does not exist")
        self.dependencies.append((set(pre),set(post)))
        
    def remove_dependency(self, pre, post):
        '''
        remove dependency from the dependency list of the scheme

        Args:
            pre(str): attributes that initiate the FD (left of the arrow)
            post(str): attributes that are determined by the FD (right of the arrow)
        '''
        for i in chain(pre, post):
            if i not in self.attributes:
                # exception when an attribute is used that is not in the list of attributes of the dependency
                raise Exception(f"Attribute {i} does not exist")
        self.dependencies.remove((set(pre), set(post)))


    def get_attr_closure(self, attr):
        '''
        get the close of the given attribute
        
        Args:
            attr(str): the name of the attribute to calculate the closure for
            
        Returns:
            set: the closure of the attribute 
        '''
        #closure set is build up iteratively, until it does not expand anymore
        closure = set(attr)
        #set of previous iteration
        last = set()
        while closure != last:
            last = closure.copy()
            #check all FDs whether their initiators are part of the closure
            #and add closure of the respective FD to the calculating closure
            for dep in self.dependencies:
                left,right=dep
                if left.issubset(closure):
                    closure.update(right)     
        return closure
    
    def attribute_combinations(self):
        '''
        generator for keys
        '''
        for i in range(1, len(self.attributes)+1):
            for keys in combinations(self.attributes, i):
                yield keys
        
    def find_candidate_keys(self):
        '''
        find candidate keys of the scheme
        '''
        ans = []
        #check closures of all attributes and attribute combinations iteratively
        #smaller candidate keys added first
        for keys in self.attribute_combinations():
            if(self.get_attr_closure(keys) == self.attributes):
                k = set(keys)
                #no subset of currently checked key is already in
                if not any([x.issubset(k) for x in ans]):
                    ans.append(k)
        return ans
    
    def decompose(self):
        '''
        decomposition algorithm
        '''
        self.isdecomposed = True
        self.tables = [self.attributes]
        for dep in self.dependencies:
            left,right=dep
            for attr_set in self.tables:
                #newset contains the unity of attributes of the FD
                newset = left.symmetric_difference(right)
                #if newset is real subset, extra attributes still exist
                #--> need to break it up
                if newset.issubset(attr_set) and newset != attr_set:
                    ## print("Splitting {} into {} and {}".format(attr_set, attr_set.difference(dep[1]), newset))
                    #split attributes of the FD closure off the attribute set
                    attr_set.difference_update(right)

                    #add new BCNF set to list of attribute sets
                    self.tables.append(newset)
        return self.tables
        
    def decompose_all(self):
        ## Messy sets and tuples to get rid of duplicates, eew
        tables_possibilities = []
        
        for ordering in permutations(self.dependencies):
            tbl = [self.attributes.copy()]
            
            for dep in ordering:
                left,right=dep
                for attr_set in tbl:
                    newset = left.symmetric_difference(right)
                    if newset.issubset(attr_set) and newset != attr_set:
                        attr_set.difference_update(right)
                        tbl.append(newset)
                        
            tbl = [tuple(x) for x in tbl]
            tables_possibilities.append(tuple(tbl))
            
        return set(tables_possibilities)
        
    def is_lossy(self):
        '''
        check for lossyness
        
        Returns:
            bool: True if if one of my dependencies is not preserved
        '''
        if not self.isdecomposed:
            raise Exception("Can't tell if lossy if the FD hasn't been decomposed yet")
        for dep in self.dependencies:
            if not self.is_preserved(dep):
                return True
        return False
        
    def is_preserved(self, dep):
        '''
        check whether the given dependency is preserved
        
        Args:
            dep(): the dependency to check
        
        Returns:
            bool: True if 
        '''
        left,right=dep
        pre = left.symmetric_difference(right)
        for attr_set in self.tables:
            if pre == attr_set:
                return True
        return False

    def calculate_fds_in_subset(self, subset):
        '''
        calculate all dependencies in a subset. Also includes dependencies for which
        attribute parts are missing because they are not in the subset. Does not include
        original dependencies that have lost all there attributes in precondition or closure

        '''
        subset_dependencies = []
        for dep in self.dependencies:
            new_dep_pre = set()
            new_dep_post = set()
            left,right=dep
            #check whether attributes occur in pre or post of the original FD
            for attr in left:
                if attr in subset:
                    new_dep_pre.add(attr)
            for attr in right:
                if attr in subset:
                    new_dep_post.add(attr)
            #only add new dependency if none of both sides is empty
            if new_dep_pre != set() and new_dep_post != set():
                subset_dependencies.append((new_dep_pre, new_dep_post))
        return subset_dependencies


    def is2NF(self):
        '''
        calculates whether the FD set is in 2NF: Every attribute has to depend on the whole CK.
        Check for every attribute whether ther is a part of any of the CKs which has the attribute in its closure
        '''
        ckeys = self.find_candidate_keys()
        # check every non-ck-attribute
        for attr in self.attributes:
            skip = False
            for ckey in ckeys:
                for ckey_part in ckey:
                    if attr == ckey_part:
                        skip = True
                        
            if skip == True:
                continue

            # check every key candidate
            for ckey in ckeys:
                # check every subset of keys (not yet)
                for ckey_part in ckey:
                    ckey_part_closure = self.get_attr_closure(ckey_part)
                    if attr in ckey_part_closure:
                        return False
        return True
    
    def is3NF(self):
        '''
        calculates whether the FD set is in 3NF: There are no dependencies between non-key attributes
        '''
        ckeys = self.find_candidate_keys()

        for dep in self.dependencies:
            left,right=dep
            # get all attributes of an fd
            dep_attributes = set()
            dep_attributes.update(left)
            dep_attributes.update(right)
            dep_has_ckey_attr = False

            # check all attributes of the fd whether at least one of them is contained in a ckey
            for attr in dep_attributes:
                for ckey in ckeys:
                    if set(attr).issubset(ckey):
                        dep_has_ckey_attr = True
                        break
            if not dep_has_ckey_attr:
                return False
        return True
        
    def generate_cluster(self,shape:str='box',indent:str='  '):
        '''
        graphviz digraph subgraph (cluster) generation for this functional dependency set
        
        Args:
            shape(str): the shape to use - default: box
            indent(str): indentation - default: two spaces
        Return:
            str: graphviz markup
        '''
        markup=''
        # sort dependencies by largest pre
        dependencies = self.dependencies.copy()
        dependencies.sort(key=lambda dep: len(dep[0]), reverse=True)

        # collect attributes that are only on the right side
        only_post = self.attributes.copy()
        # generate clusters
        cluster_markup = ''
        for dep in dependencies:
            pre, post = dep
            only_post -= pre
            cluster_name=''.join(sorted(pre))
            cluster_markup += f'{indent}subgraph cluster_{cluster_name}{{\n'
            cluster_markup += f'{indent} label="{cluster_name}"\n'
            for attrVar in sorted(pre):
                attr = self.attribute_map[attrVar]
                cluster_markup += f'{indent}{indent}{attrVar} [shape={shape} label="{attr}"]\n'
            cluster_markup += f'{indent}}}\n'
        
        # generate arrows
        arrow_markup = ''
        for dep in dependencies:
            pre, post = dep
            for attrVar in sorted(post):
                arrow_markup += f'{indent}{sorted(pre)[0]}->{attrVar}\n'

        # create markup for only post attributes
        only_post_markup = ''
        for attrVar in sorted(only_post):
            attr = self.attribute_map[attrVar]
            only_post_markup += f'{indent}{attrVar} [shape={shape} label="{attr}"]\n'

        # concatenate markup
        markup += only_post_markup
        markup += cluster_markup
        markup += arrow_markup
        return markup

    def as_graphviz(self,withCluster:bool=True):
        '''
        
        convert me to a graphviz markup e.g. to try out in 
        
        http://magjac.com/graphviz-visual-editor/
        or 
        http://diagrams.bitplan.com
        
        Return:
            str: the graphviz markup for this functional DependencySet
        '''
        markup=f"#generated by {__file__} on {self.isodate}\n"
        markup+="digraph functionalDependencySet{"
        # add title see https://stackoverflow.com/a/6452088/1497139
        markup+=f'''
  // title
  labelloc="t";
  label="{self.title}"
'''
        if not withCluster:
            markup+="// Attribute variables \n"   
            for attrVar in sorted(self.attributes):
                attr=self.attribute_map[attrVar]
                markup+=(f"""  {attrVar} [ shape=box label="{attr}"] \n""")
        else:
            markup+=self.generate_cluster()
        markup+="}"
        return markup

    def left_reduction(self):
        '''
        executes a left reduction on the dependencies from this fdset
        '''
        remaining_deps = self.dependencies.copy()
        while remaining_deps:
            dep = remaining_deps.pop(0)
            pre, post = dep
            for attr in sorted(pre):
                if post <= self.get_attr_closure(pre - {attr}):
                    self.remove_dependency(pre, post)
                    self.add_dependency(pre - {attr}, post)
                    pre = pre - {attr}

    def right_reduction(self):
        '''
        executes a right reduction on the dependencies from this fdset
        '''
        remaining_deps = self.dependencies.copy()
        while remaining_deps:
            dep = remaining_deps.pop(0)
            pre, post = dep
            for attr in sorted(post):
                self.remove_dependency(pre, post)
                self.add_dependency(pre, post - {attr})
                if {attr} <= set(self.get_attr_closure(pre)):
                    post = post - {attr}
                else:
                    self.remove_dependency(pre, post - {attr})
                    self.add_dependency(pre, post)

    def remove_empty_fds(self):
        '''
        remove empty fds of form "A → {}" from this fdset
        '''
        for dep in self.dependencies:
            pre, post = dep
            if post == set():
                self.remove_dependency(pre, post)

    def combine_fds(self):
        '''
        executes a left reduction for this fdset
        '''
        combined_dependencies = []
        while self.dependencies:
            pre, post = self.dependencies.pop(0)
            new_post = post
            for dep in self.dependencies:
                left,right=dep
                if left == pre:
                    new_post = new_post | right
                    self.remove_dependency(left, right)
            combined_dependencies.append((pre, new_post)) 
        self.dependencies = combined_dependencies

    def canonical_cover(self):
        '''
        determines the canonical cover of this fdset

        4 substeps with respective functions

        https://git.rwth-aachen.de/i5/teaching/dbis-vl/-/raw/main/6-RelDesign/6-RelationaleEntwurfstheorie.pdf#page=39
        '''
        self.left_reduction()
        self.right_reduction()
        self.remove_empty_fds()
        self.combine_fds()

    def create_new_fdsets(self):
        '''
        create fdsets from the dependencies resulting from the canonical cover

        Return:
            list[FunctionalDependencySet]: list of fdsets created from the dependencies resulting from the canonical cover
        '''
        deps = self.dependencies.copy()
        i = 1
        new_fdsets = []
        while deps:
            tmp = deps.pop(0)
            pre, post = tmp
            new_attributes = pre | post
            new_deps = [tmp]
            for dep in deps:
                left,right=dep
                if left | right <= new_attributes:
                    new_deps.append(dep)
                    deps.remove(dep)
            fds = FunctionalDependencySet(new_attributes, 'R' + str(i))
            i += 1
            for dep in new_deps:
                left,right=dep
                fds.add_dependency(left,right)
            new_fdsets.append(fds)
        return new_fdsets

    def synthesize(self):
        '''
        synthesize algorithm
        
        see https://git.rwth-aachen.de/i5/teaching/dbis-vl/-/raw/main/6-RelDesign/6-RelationaleEntwurfstheorie.pdf#page=76
        and Kemper page 197

        Return:
            list[FunctionalDependencySet]: list of synthesized fdsets deriving from this fdset
        '''
        keys = self.find_candidate_keys()
        self.canonical_cover()
        fdsets = self.create_new_fdsets()
        fdsets_with_key = self.create_optional_key_scheme(keys, fdsets)
        reduced_fdsets = self.remove_subset_relations(fdsets_with_key)
        return reduced_fdsets

    def create_optional_key_scheme(self, keys, fdsets):
        '''
        creates a new fdset if key is not subset of any of the existing sets attributes

        Return:
            list[FunctionalDependencySet]: The list of fdsets with relation that has key candidate of original scheme
        '''
        for key in keys:
            for fds in fdsets:
                if set(key) <= fds.attributes:
                    return fdsets
        key = set(keys[0])
        fds = FunctionalDependencySet(key, 'R' + str(len(fdsets) + 1))
        m = len(key)//2
        # fds.add_dependency(sorted(key)[:m], sorted(key)[m:])
        fds.add_dependency(key, key)
        fdsets.append(fds)
        return fdsets

    def remove_subset_relations(self, fdsets):
        '''
        removes fdsets with attributes that are a subset of another fdset

        Return:
            list[FunctionalDependencySet]: The reduced list of fdsets
        '''
        if self.debug:
            print(fdsets)
        for fds in fdsets.copy():
            attributes = fds.attributes
            conflict = next((fdset for fdset in fdsets if fds.title != fdset.title and attributes <= fdset.attributes), None)
            if conflict is not None:
                fdsets.remove(fds)
        return fdsets




'''
Created on 2022-06-08

@author: wf
'''
from enum import Enum

class Notation(str,Enum):
    '''
    a notation to be used for stringifcation
    '''
    math="LaTex math symbols"   
    utf8="UTF-8 Symbols"  
    plain="plaintext ASCII"
    short="short notation without delimiters using juxtapositions" 
    
class Set(set):
    '''
    https://docs.python.org/3/tutorial/datastructures.html#sets
    '''
    notation=Notation.plain
    
    def __init__(self,*args, **kwargs):
        '''
        constructor
        '''
        super().__init__(*args,**kwargs)
        
    def __str__(self):
        text=Set.stringify_set(self,notation=Set.notation)
        return text
    
    @classmethod    
    def stringify_set(cls,pSet,notation:Notation):
        '''
        return a string representation of the given set using the given delimiter
        
        Args:
            pSet(set): the set to stringify
            delim(str): the delimiter to use
        '''
        sortedElements=sorted([str(elem) for elem in pSet])
        elementDelim=","
        if notation==Notation.short:
            elementDelim='' # Thin space would be better
        if notation==Notation.short:
            text=''    
        elif notation==Notation.math:
            text="\{"
        else:
            text="{"

        delim=""
        for element in sortedElements:
            text+=f"{delim}{element}"
            delim=elementDelim
        if notation==Notation.short:
            pass 
        elif notation==Notation.math:
            text+="\}"
        else:
            text+="}"      
        return text
    
class FD(object):
    """A functional dependency with left- and right-hand side."""
    notation=Notation.plain

    def __init__(self, left, right):
        """Create FD with left hand side  and right hand side
        
        Args:
            left(object): set of attributes for the left hand side
            right(object): set of attributes for the right hand side 

        """
        self.left=Set(left)
        self.right=Set(right)
        
    def __str__(self):
        '''
        convert me to a string
        
        Return:
            str: a string representation of myself
        '''
        text=FD.stringify_FD(self, FD.notation)
        return text
    
    @classmethod    
    def stringify_FD(cls,fd,notation:Notation):
        '''
        Return:
            a string representation of the given Functional Dependency
        '''
        setNotation=Notation.short
        leftText=Set.stringify_set(fd.left, notation=setNotation)
        rightText=Set.stringify_set(fd.right,notation=setNotation)
        if notation==Notation.utf8:
            to="→"
        elif notation==Notation.math:
            to="\\to "
        else:
            to="->"
        text=f"{leftText}{to}{rightText}"
        return text
    
class Attribute:
    '''
    an Attribute e.g. 
        Example: Attribute('A', 'Wikidata identifier', 'Wikidata-Schlüssel')
    '''
    def __init__(self, var_name:str, english_name:str, german_name:str):
        '''
        constructor

        Args:
            var_name(str): the Variable name
            english_name(str): the english name
            german_name(str): the german name
 
        '''
        self.var_name=var_name
        self.german_name=german_name
        self.english_name=english_name
        
    def __str__(self):
        text=f"{self.var_name}≡{self.english_name}≡{self.german_name}"
        return text
    
class RelSchema(object):
    """A relation schema consists of a set of attributes and a set of FDs.

    Various normal forms exist to describe "good" schemata.  Normalization
    is the process of creating schemata that satisfy certain normal forms.
    The class of synthesis algorithms targets 3NF.
    """

    def __init__(self, attributes:Set, fds,notation:Notation=None):
        """
        Construct relation schema with attributes and FDs.
        
        Args:
            attributes(list): a list of Attributes
            fds(FunctionalDependencySet): a set of FDs
        """
        if notation is None:
            notation=Notation.utf8
        self.notation=notation
        self.attributes = attributes
        self.fds = fds
        
    def __str__(self):
        text=RelSchema.stringify(self, self.notation)
        return text
        
    @classmethod
    def stringify(cls,rs,notation):
        '''
        return a textual representation of the given relational schema in the given notation
        '''
        text="R={"
        text+=Set.stringify_set(rs.attributes, notation)
        text+=f",{rs.fds}"
        text+="}"
        return text
           


import pathlib
from setuptools import setup

class Version(object):
    name="relational_model"
    description="RWTH Aachen Computer Science i5/dbis assets for Lecture Datenbanken und Informationssysteme"
    version='0.0.4'

# The directory containing this file
HERE = pathlib.Path(__file__).parent.resolve()

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="dbis-functional-dependencies",
    version=Version.version,
    description=Version.description,
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://git.rwth-aachen.de/i5/teaching/dbis-functional-dependencies",
    author="Wolfgang Fahl",
    author_email="fahl@dbis.rwth-aachen.de",
    license="Apache",
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    packages=["functional_dependencies"],
    include_package_data=True,
    install_requires=[]
)
